/*
 *   Copyright 2012 Institute of Communication and Computer Systems (ICCS) 
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 *
 */

/**
 * @file   dmmlib.c
 * @author Ioannis Koutras (joko@microlab.ntua.gr)
 * @date   September 2012
 *
 * @brief  Implementations of malloc(), free() and calloc() calls.
 */

#include "dmmlib/dmmlib.h"
#ifdef FL_RB_ONLY
#include "freelist/freelist_malloc.h"
#endif /* FL_RB_ONLY */
#ifdef BITMAP_RB_ONLY
#include "bitmap/bitmap.h"
#include "bitmap/bitmap_rb.h"
#endif /* BITMAP_RB_ONLY */
#include <stdbool.h>
#include <string.h> /* for memset() */

void * malloc(size_t size) {
    raw_block_header_t *raw_block, *new_raw_block;
    size_t allocation_size;
    void *ptr;

    raw_block = systemallocator.raw_block_head;
    ptr = NULL;

    /* Try to find a raw block available for allocation */

    while(raw_block != NULL) {
        ptr = dmmlib_malloc(raw_block, size);
        if(ptr != NULL) {
            break;
        }
        raw_block = raw_block->next_raw_block;
    }

    if(ptr == NULL) {
#ifdef FL_RB_ONLY
        allocation_size = size + sizeof(raw_block_header_t) +
            sizeof(freelist_rb_t);
#endif /* FL_RB_ONLY */
#ifdef BITMAP_RB_ONLY
        allocation_size = size + sizeof(raw_block_header_t) +
            sizeof(bitmap_rb_t);
#endif /* BITMAP_RB_ONLY */

        if(allocation_size < SYS_ALLOC_SIZE) {
            allocation_size = SYS_ALLOC_SIZE;
        } else {
            /* If allocation size is more than the default size, then request
             * multiples of it.
             */
            allocation_size = (allocation_size / SYS_ALLOC_SIZE + 1) *
                SYS_ALLOC_SIZE;
        }

        pthread_mutex_lock(&systemallocator.creation_mutex);
        new_raw_block = create_new_raw_block(allocation_size);
        if(new_raw_block != NULL) {
            new_raw_block->next_raw_block = systemallocator.raw_block_head;
            systemallocator.raw_block_head = new_raw_block;
            pthread_mutex_unlock(&systemallocator.creation_mutex);
            ptr = dmmlib_malloc(new_raw_block, size);
        }
    }
    
    return ptr;
}

void free(void *ptr) {
    raw_block_header_t *current_raw_block;
    bool found;

    if(ptr == NULL) {
        return;
    }

    found = false;

    current_raw_block = systemallocator.raw_block_head;
    while(current_raw_block) {
        if((char *)ptr - (char *)(current_raw_block) -
                sizeof(raw_block_header_t) < current_raw_block->size) {
            found = true;
            break;
        }
        current_raw_block = current_raw_block->next_raw_block;
    }

    if(found == true) {
        dmmlib_free(current_raw_block, ptr);
    }
}

void * calloc(size_t nmemb, size_t size) {
    size_t i;
    char *buf[nmemb];

    if(nmemb == 0 || size == 0) {
        return NULL;
    }

    i = 0;

    while(i < nmemb) {
        buf[i] = (char*)malloc(size);
        memset(buf[i], '\0', size);
        i++;
    }

    return buf[0];
}
